#!/usr/bin/env python

"""
Access library for sqlite. This layer of abstraction is to help in 
swapping out sqlite for another database such as mysql. It does *not*
do anything for sql quirks.

init()
execute()
query()
commit()
close()


"""
## std libs
import time, thread, random

## 3rd-party libs
import sqlite

## project libs
import util

cx = None	##connection obj
initialized = 0

"""
initializes the database connection; creates tables and indexes if necessary
"""
def init():
	global cx, initialized
	if not initialized:
		dbfile = util.home + "/dowser/library.db"
		
		util.mkdir(util.home + "/dowser")
			
		cx = sqlite.connect(dbfile)
		
		make_schema()
		initialized = 1
	else:
		pass ##util.debug( "db initialized already." )

"""
executes a SELECT statement and returns a rowset as an array of dictionaries
todo: use of lock()/release() makes read access to the db single-threaded.
"""	
def query(sql):
	lock()
	try:
		global cx
		cur = cx.cursor()
		cur.execute(sql)
		release()
		return cur.fetchall()
	except:
		release()
		raise
	

"""
executes INSERTS, UPDATES and other statements that alter data
and returns either the last row id (INSERT) or number of rows
affected (UPDATE, etc)

todo: need rows affected code. separate update() function?
"""	
def execute(sql, autocommit=True):
	lock()
	try:
		global cx
		cur = cx.cursor()
		cur.execute(sql)
		if (autocommit):
			commit()
		release()
		return cur.lastrowid
	except:
		release()
		raise

"""
Commits a transaction
nb: does *not* use the lock mechanism.
"""
def commit():
	global cx
	cx.commit()

"""
simple "deli" or "FIFO" queue for the database object. The thread grabs 
the next "ticket" number and waits until it comes up. Since this is 
python, we can get away with an unlimited number of "tickets"
without overflowing an integer down the road. 

NB: you MUST be VERY careful to release your locks when an exception
is thrown, else you'll have threads waiting in the queue forever.

todo: what happens when two threads release() right after each other?
could a thread get skipped?
"""
next = 0
current = 0
def lock():
	global next, current
	mynum = next
	next = next + 1
	##util.debug( "my number is %d. next is %d" % (mynum, next) )
	while mynum != current:
		pass
	
def release():
	global current
	current = current + 1





		
"""
Checks the database to make sure all the tables exist.
XXX: if you make a change to the schema, make sure it is "additive".
changing an existing table (adding a column, whatever) is
going to need additional code here to "upgrade" old installs. 
Assume you have 10k users with databases they don't want to lose.
"""
def make_schema():
	indexes = {
		'kw_idx': 
			'create unique index kw_idx on keywords (keyword);',
		'kw_hash_idx':
			'create index kw_hash_idx on keywords (kwhash);',
		'url_idx':
			'create unique index url_idx on url (url);',
		'url_hash_idx':
			'create index url_hash_idx on url (urlhash);',
		'url_path_idx':
			'create index url_path_idx on url (cachepath);',
		'dh_idx':
			'create index dh_idx on url (datahash);',
		'domain_idx':
			'create index domain_idx on url (domain);',
		'crawl_idx':
			'create index crawl_idx on url (lastCrawled);',
		'cache_idx':
			'create index cache_idx on cache_history (urlhash);',
		'map_kw_url_idx':
			'create unique index map_kw_url_idx on map_keyword_url (kwHash,urlHash);',
		'map_kw_idx':
			'create index map_kw_idx on map_keyword_url (kwHash);',
		'map_url_idx': 
			'create index map_url_idx on map_keyword_url (urlHash);',
		'map_kw_search_idx': 
			 'create unique index map_kw_search_idx on map_search_kw (sid,kwHash);',
		'map_kw_kw_idx': 
			 'create unique index map_kw_kw_idx on map_kw_kw (sid,kw,kw2);',
		'map_kw_kw_idx2': 
			 'create index map_kw_kw_idx2 on map_kw_kw (kw);',
		'map_kw_kw_idx3': 
			 'create index map_kw_kw_idx3 on map_kw_kw (kw2);'
	}
	
		
	tables=	{
		'keywords':"""
			create table keywords (
				kwid integer primary key,
				kwHash char(40) not null,
				keyword varchar(255) not null,
				inRange int default 0 not null
			)
			""", 
		'url':"""
			create table url (
				urlid integer primary key,
				urlHash char(40) not null,
				url clob not null,
				datahash varchar(40),
				cachepath clob,
				domain varchar(512) not null,
				title varchar(255),
				descr varchar(255),
				size int,
				lastCrawled int,
				mimetype varchar(255) default 'text/html',
				status int default 200,
				lastsuccess int,
				retryafter int default 15,
				lastused int
			)""",
		'cache_history':"""
			create table cache_history (
				urlhash char (40) not null,
				cachedate int not null,
				datahash char (40) not null,
				cachepath clob not null
			)""", 
		'map_keyword_url':"""
			create table map_keyword_url (
				kwHash char(40) not null,
				urlHash char(40) not null,
				count integer,
				snippet varchar(255)
			)""", 
		'searches':"""
			create table searches (
				sid integer primary key,
				sdate integer not null,
				keywords varchar(255) not null
			)""", 
		'map_search_kw':"""
			create table map_search_kw (
				sid integer not null,
				kwHash char(40) not null,
				idx integer not null
			)""", 
		'map_kw_kw':"""
			create table map_kw_kw (
				kw char(40) not null,
				kw2 char(40) not null,
				sid integer not null
			)"""

	}
	
	##check for & add tables
	rows = query("select name from sqlite_master where type ='table' order by name")
	tbl = {}
	for row in rows:
		tbl.update({row.name:True})
	for k,v in tables.iteritems():
		if not tbl.has_key(k):
			execute(v, False)
	commit()


	###check indexes and add any that aren't there
	rows = query("select name from sqlite_master where type ='index' order by name")
	idx = {}
	for row in rows:
		idx.update({row.name:True})
	for k,v in indexes.iteritems():
		if not idx.has_key(k):
			execute(v, False)
	commit()


def close():
	global cx, initialized
	cx.close()
	initialized = 0
	
